from __future__ import print_function, division

import re

import numpy as np

from astroquery.vizier import Vizier

from astropy.coordinates import SkyCoord
from astropy.table import Table
import astropy.units as units

from itertools import izip

from ccdproc import CCDData
__all__ = [

    'in_frame',
    'catalog_search',
    'catalog_clean',
]


def scale_and_downsample(data, downsample=4, min_percent=20, max_percent=99.5):
    if downsample > 1:
        scaled_data = block_reduce(scaled_data, block_size=(downsample, downsample))
    return scaled_data


def in_frame(frame_wcs, coordinates):
    """
    Description:Check which of a set of coordinates are in the footprint of
    the WCS of an image
    Preconditions:
    Postconditions:
    """
    x, y = frame_wcs.all_world2pix(coordinates.ra, coordinates.dec, 0)
    in_x = (x >= 0) & (x <= frame_wcs._naxis1)
    in_y = (y >= 0) & (y <= frame_wcs._naxis2)
    return in_x & in_y


def catalog_search(frame_wcs, shape, desired_catalog,
                   ra_column='RAJ2000',
                   dec_column='DEJ2000',
                   radius=0.5,
                   clip_by_frame=True):
    """
    Description: This function takes coordinate data from an image and a
    catalog name and returns the positions of those stars.
    Preconditions:frame_wcs is a WCS object, shape is tuple, list or array of
    numerical values, desired_catalog is a string and radius is a numerical
    value.
    Postconditions:
    """
    rad = radius * units.deg
    # Find the center of the frame
    center_coord = frame_wcs.all_pix2world([[shape[1] / 2, shape[0] / 2]], 0)
    center = SkyCoord(center_coord, frame='icrs', unit='deg')

    # Get catalog via cone search
    Vizier.ROW_LIMIT = -1  # Set row_limit to have no limit
    cat = Vizier.query_region(center, radius=rad, catalog=desired_catalog)
    # Vizier always returns list even if there is only one element. Grab that
    # element.
    cat = cat[0]
    cat_coords = SkyCoord(ra=cat[ra_column], dec=cat[dec_column])
    if clip_by_frame:
        in_fov = in_frame(frame_wcs, cat_coords)
    else:
        in_fov = np.ones([len(cat_coords)], dtype=np.bool)
    x, y = frame_wcs.all_world2pix(cat_coords.ra, cat_coords.dec, 0)
    return (cat[in_fov], x[in_fov], y[in_fov])


def catalog_clean(catalog, remove_rows_with_mask=True,
                  **other_restrictions):
    """
    Return a catalog with only the rows that meet the criteria specified.

    Parameters
    ----------

    catalog : astropy.Table
        Table of catalog information. There are no restrictions on the columns.
    remove_rows_with_mask : bool, optional
        If ``True``, remove rows in which one or more of the values is masked.
    other_restrictions: dict, optional
        Key/value pairs in which the key is the name of a column in the
        catalog and the value is the criteria that values in that column
        must satisfy to be kept in the cleaned catalog. The criteria must be
        simple, beginning with a comparison operator and including a value.
        See Examples below.
    """

    comparisons = {
        '<': np.less,
        '=': np.equal,
        '>': np.greater,
        '<=': np.less_equal,
        '>=': np.greater_equal,
        '!=': np.not_equal
    }

    recognized_comparison_ops = '|'.join(comparisons.keys())
    keepers = np.ones([len(catalog)], dtype=np.bool)

    if remove_rows_with_mask and catalog.masked:
        for c in catalog.columns:
            keepers &= ~catalog[c].mask

    for column, restriction in other_restrictions.iteritems():
        criteria_re = re.compile(r'({})(.+)'.format(recognized_comparison_ops))
        results = criteria_re.match(restriction)
        if not results:
            raise ValueError("Criteria {}{} not "
                             "understood.".format(column, restriction))
        comparison_func = comparisons[results.group(1)]
        comparison_value = results.group(2)
        new_keepers = comparison_func(catalog[column], np.float(comparison_value))
        keepers = keepers & new_keepers

    return catalog[keepers]



from __future__ import print_function, division

import numpy as np
import pytest

from ..catalog_search import catalog_clean

from astropy.table import Table


def a_table(masked=False):
    test_table = Table([(1, 2, 3), (1, -1, -1)], names=('a', 'b'),
                       masked=masked)
    return test_table


def test_clean_criteria_none_removed():
    """
    If all rows satisfy the criteria, none should be removed.
    """
    inp = a_table()
    criteria = {'a': '>0'}
    out = catalog_clean(inp, **criteria)
    assert len(out) == len(inp)
    assert (out == inp).all()


@pytest.mark.parametrize("condition",
                         ['>0', '=1', '!=-1', '>=1'])
def test_clean_criteria_some_removed(condition):
    """
    Try a few filters which removes the second row and check that it is
    removed.
    """
    inp = a_table()
    criteria = {'b': condition}
    out = catalog_clean(inp, **criteria)
    assert len(out) == 1
    assert (out[0] == inp[0]).all()


@pytest.mark.parametrize("clean_masked",
                         [False, True])
def test_clean_masked_handled_correctly(clean_masked):
    inp = a_table(masked=True)
    # Mask negative values
    inp['b'].mask = inp['b'] < 0
    out = catalog_clean(inp, remove_rows_with_mask=clean_masked)
    if clean_masked:
        assert len(out) == 1
        assert (np.array(out[0]) == np.array(inp[0])).all()
    else:
        assert len(out) == len(inp)
        assert (out == inp).all()


def test_clean_masked_and_criteria():
    """
    Check whether removing masked rows and using a criteria work
    together.
    """
    inp = a_table(masked=True)
    # Mask the first row.
    inp['b'].mask = inp['b'] > 0

    inp_copy = inp.copy()
    # This should remove the third row.
    criteria = {'a': '<=2'}

    out = catalog_clean(inp, remove_rows_with_mask=True, **criteria)

    # Is only one row left?
    assert len(out) == 1

    # Is the row that is left the same as the second row of the input?
    assert (np.array(out[0]) == np.array(inp[1])).all()

    # Is the input table unchanged?
    assert (inp == inp_copy).all()


@pytest.mark.parametrize("criteria,error_msg", [
                         ({'a': '5'}, "not understood"),
                         ({'a': '<foo'}, "could not convert string")])
def test_clean_bad_criteria(criteria, error_msg):
    """
    Make sure the appropriate error is raised when bad criteria are used.
    """
    inp = a_table(masked=False)

    with pytest.raises(ValueError) as e:
        catalog_clean(inp, **criteria)
    assert error_msg in str(e)



from __future__ import print_function, division

import numpy as np
from scipy import optimize

import astropy.units as u
from astropy.table import Table
from astropy.coordinates import SkyCoord

__all__ = [
    'filter_transform',
    'standard_magnitude_transform',
    'calculate_transform_coefficients',
]


def filter_transform(mag_data, output_filter, R=None, B=None,
                     V=None, I=None, g=None, r=None, i=None):
    '''
    Transform SDSS magnitudes to BVRI using Ivezic et all (2007).

    Description: This function impliments the transforms in 'A Comparison of SDSS Standard
    Preconditions: mag_data must be an astropy.table object consisting of numerical values,
    output_filter must be a string 'R', 'B', 'V', or 'I' and for any output filter must be passed a
    corresponding key (arguemnts R, B, V...) to access the necissary filter information from
    mag_data
    Postconditions: returns a

    # #Basic filter transforms from Ivezic et all (2007)

    '''
    if output_filter == 'R':
        if r and i is not None:
            try:
                r_mags = mag_data[r]
            except:
                raise KeyError('key', str(
                    r), 'not found in mag data for r mags')
            try:
                i_mags = mag_data[i]
            except:
                raise KeyError('key', str(
                    i), 'not found in mag data for i mags')
            A = -0.0107
            B = 0.0050
            C = -0.2689
            D = -0.1540
            c = r_mags - i_mags
            R_mag = (A * (c**3)) + (B * (c**2)) + (C * c) + D + r_mags
            R_mag.name = 'R_mag'
            R_mag.description = 'R-band magnitude transformed from r-band and i-band'
            return R_mag
        else:
            raise KeyError(
                'arguemnts r and i must be defined to transform to I filter')

    if output_filter == 'I':
        if r and i is not None:
            try:
                r_mags = mag_data[r]
            except KeyError:
                raise KeyError('key', str(
                    r), 'not found in mag data for r mags')
            try:
                i_mags = mag_data[i]
            except KeyError:
                raise KeyError('key', str(
                    i), 'not found in mag data for i mags')
            A = -0.0307
            B = 0.1163
            C = -0.3341
            D = -0.3584
            c = r_mags - i_mags
            I_mag = (A * (c**3)) + (B * (c**2)) + (C * c) + D + r_mags
            I_mag.name = 'I_mag'
            I_mag.description = 'I-band magnitude transformed from r-band and i-band'
            return I_mag
        else:
            raise KeyError(
                'arguments r and i must be defined to transform to I filter')

    if output_filter == 'B':
        if r and g is not None:
            try:
                r_mags = mag_data[r]
            except KeyError:
                raise KeyError('key', str(
                    r), 'not found in mag data for r mags')
            try:
                g_mags = mag_data[g]
            except KeyError:
                raise KeyError('key', str(
                    i), 'not found in mag data for g mags')
            A = 0.2628
            B = -0.7952
            C = 1.0544
            D = 0.02684
            c = g_mags - r_mags
            B_mag = (A * (c**3)) + (B * (c**2)) + (C * c) + D + r_mags
            B_mag.name = 'B_mag'
            B_mag.description = 'B-band magnitude transformed from r-band and g-band'
            return B_mag
        else:
            raise KeyError(
                'arguemnts r and g must be defined to transform to B filter')

    if output_filter == 'V':
        if r and g is not None:
            try:
                r_mags = mag_data[r]
            except KeyError:
                raise KeyError('key', str(
                    r), 'not found in mag data for r mags')
            try:
                g_mags = mag_data[g]
            except KeyError:
                raise KeyError('key', str(
                    i), 'not found in mag data for g mags')
            A = 0.0688
            B = -0.2056
            C = -0.3838
            D = -0.0534
            c = g_mags - r_mags
            V_mag = (A * (c**3)) + (B * (c**2)) + (C * c) + D + r_mags
            V_mag.name = 'V_mag'
            V_mag.description = 'V-band magnitude transformed from r-band and g-band'
            return V_mag
        else:
            raise KeyError(
                'arguments r and g must be defined to transform to B filter')
    else:
        raise ValueError('the desired filter must be a string R B V or I')


# Define the log-likelihood via the Huber loss function
def huber_loss(m, b, x, y, dy, c=2):
    y_fit = m * x + b
    t = abs((y - y_fit) / dy)
    flag = t > c
    return np.sum((~flag) * (0.5 * t ** 2) - (flag) * c * (0.5 * c - t), -1)


def standard_magnitude_transform(instrumental,
                                 catalog,
                                 for_filter=None,
                                 match_radius=0.5 * u.arcsec,
                                 color_from='catalog',
                                 color_1='B',
                                 color_2='V',):
    """
    Given catalog magnitudes and instrumental magnitudes, calculate transform
    from instrumental to catalog. Colors can be drawn from the catalog or
    from instrumental magnitudes.

    Parameters
    ----------

    instrumental: astropy Table
        Table containing instrumental magnitudes from which transform
        coefficients are to be determined. Each filter should be a column
        whose name is the filter. In addition, there should be a column
        called 'RA' and a column called 'Dec'.
    catalog : astropy Table
        Table containing catalog magnitudes from which transform
        coefficients are to be determined. Each filter should be a column
        whose name is the filter. In addition there should be a column
        called 'RA' and a column called 'Dec'.
    for_filter : str
        Filter for which to calculate the transform. Filter name must be a
        column name in instrumental and catalog.
    match_radius : astropy Quantity  or float
        Determines cutoff for matching coordinates of instrumental
        magnitudes to catalog magnitudes. An on-sky separation of
        less than match_radius counts as a match. If no units are provided
        it is assumed the unit is degrees.
    color_from : str
        Either 'catalog' or 'instrumental', indicates which catalog should be
        used for calculating the color in the transform.
    color_1 : str
        First filter to use for calculating color.
    color_2 : str
        Second color to use for calculating the color.

    """
    tables = {
        'catalog': catalog,
        'instrumental': instrumental
    }

    if not for_filter:
        raise ValueError('Must provide a value for for_filter.')

    for name, table in tables.iteritems():
        if for_filter not in table.colnames:
            raise ValueError('Filter {} not found in {} '
                             'table'.format(for_filter, name))

    instrumental_coords = SkyCoord(ra=instrumental['RA'],
                                   dec=instrumental['Dec'])
    catalog_coords = SkyCoord(ra=catalog['RA'], dec=catalog['Dec'])

    catalog_index, d2d, d3d = \
        instrumental_coords.match_to_catalog_sky(catalog_coords)

    good_match = np.array(d2d < match_radius)

    if not good_match.sum():
        raise ValueError('No matches found between instrumental'
                         ' and catalog tables.')
    if color_from == 'catalog':
        indexes = catalog_index[good_match]
    else:
        indexes = good_match
    color = tables[color_from][color_1][indexes] - tables[color_from][color_2][indexes]

    transforms = {}

    mag_diff = catalog[for_filter][catalog_index[good_match]] - instrumental[for_filter][good_match]

    # dy is error...which is simply added in quadrature, if present
    try:
        catalog_error = catalog['e_' + for_filter]
    except KeyError:
        catalog_error = np.zeros_like(good_match)

    try:
        instrumental_error = instrumental['e_' + for_filter]
    except KeyError:
        instrumental_error = np.zeros_like(good_match)
    dy = np.sqrt(catalog_error[good_match] ** 2 +
                 instrumental_error[good_match] ** 2)

    def f_huber(beta):
        return huber_loss(beta[0], beta[1],
                          x=np.array(color),
                          y=np.array(mag_diff),
                          dy=dy,
                          c=1)

    beta0 = (0.1, 20)
    beta_huber = optimize.minimize(f_huber, beta0, method='Nelder-Mead')
    # This grabs the solution the optimizer found, or raises an error
    # if no solution is found.
    if beta_huber.success:
        transforms[for_filter] = beta_huber.x
    else:
        raise RuntimeError('Optimizer did not converge')

    return transforms


def calculate_transform_coefficients(input_mag, catalog_mag, color,
                                     input_mag_error=None,
                                     catalog_mag_error=None):
    """
    Calculate linear transform coefficients from input magnitudes to catalog
    magnitudes.

    Parameters
    ----------

    input_mag : numpy array or astropy Table column
        Input magnitudes; for example, instrumental magnitudes.
    catalog_mag : numpy array or astropy Table column
        Catalog (or reference) magnitudes; the magnitudes to which the
        input_mag will eventually be transformed.
    color : numpy array or astropy Table column
        Colors to use in determining transform coefficients.
    input_mag_error : numpy array or astropy Table column, optional
        Error in input magnitudes. Default is zero.
    catalog_mag_error : numpy array or astropy Table column, optional
        Error in catalog magnitudes. Default is zero.

    Returns
    -------

    zero_point, slope : float
        Zero point and linear color term in transform equation (see notes
        below).

    Notes
    -----

    This function has some pretty serious limitations right now:

    + Errors in the independent variable are ignored.
    + Outliers are rejected using a modified loss function (Huber loss) that
      cannot be modified.
    + No errors are estimated in the calculated transformation coefficients.

    And there is all the stuff that is not listed here...
    """

    if input_mag_error is None:
        input_mag_error = np.zeros_like(input_mag)
    if catalog_mag_error is None:
        catalog_mag_error = np.zeros_like(catalog_mag)

    # Independent variable is the color, dependent variable is the
    # difference between the measured (input) magnitude and the catalog
    # magnitude.

    mag_diff = catalog_mag - input_mag

    # The error that is the errors of those combined in quadrature.
    combined_error = np.sqrt(input_mag_error**2 + catalog_mag_error**2)

    # If both errors are zero then the error is omitted.
    if (combined_error == 0).all():
        dy = None
    else:
        dy = combined_error

    def f_huber(beta):
        return huber_loss(beta[0], beta[1],
                          x=np.array(color),
                          y=np.array(mag_diff),
                          dy=dy,
                          c=1)

    # Fairly arbitrary initial guess.
    beta0 = (0.1, 20)

    beta_huber = optimize.minimize(f_huber, beta0, method='Nelder-Mead')
    # This grabs the solution the optimizer found, or raises an error
    # if no solution is found.
    if beta_huber.success:
        transforms = beta_huber.x
    else:
        raise RuntimeError('Optimizer did not converge')

    return transforms



from ccdproc import CCDData
from astropy.stats import sigma_clipped_stats
from photutils import daofind

__all__ = ['source_detection']


def source_detection(ccd, fwhm=3.0, sigma=3.0, iters=5, threshold=5.0):
    """
    Returns an astropy table containing the position of sources within the image.

    Parameters
    ----------------

    ccd : numpy.ndarray
        The CCD Image array.

    fwhm : float, optional
        Full-width half-max of stars in the image.

    sigma : float, optional.
        The number of standard deviations to use as the lower and upper clipping limit.

    iters : int, optional
        The number of iterations to perform sigma clipping
    
    threshold : float, optional.
        The absolute image value above which to select sources.
    
    Returns
    -----------

    sources
        an astropy table of the positions of sources in the image.
    """
    data = ccd.data
    mean, median, std = sigma_clipped_stats(data, sigma=sigma, iters=iters)
    sources = daofind(data - median, fwhm=fwhm, threshold=threshold*std)
    return sources



from __future__ import print_function, division, absolute_import
import numpy as np
from astropy.table import Table, join
from photutils import aperture_photometry, CircularAperture, CircularAnnulus
from .coordinates import convert_pixel_wcs

__all__ = ['photutils_stellar_photometry']

def photutils_stellar_photometry(ccd_image, sources, aperture_radius, inner_annulus,
                                 outer_annulus, gain=1.0, N_R=0, N_dark_pp=0):
    """
    Perform aperture photometry on an image, with a few options for estimating
    the local background from an annulus around the aperture.

    Parameters
    ----------
    ccd_image : `~ccdproc.CCDData`
        Image on which to perform aperture photometry.

    sources : `~astropy.table.Table`
        Table of extracted sources. Assumed to be the output of
        `~photutils.daofind()` source extraction function.

    aperture_radius : float
        Radius of aperture(s) in pixels.

    inner_annulus : float
        Inner radius of the annulus in pixels.

    outer_annulus : float
        Outer radius of the annulus in pixels.

    gain : float
        Gain of the CCD. In units of electrons per DN.

    N_R : float
        Read noise of the CCD in electrons per pixel.

    N_dark_pp : float
        Number of dark counts per pixel.

    Returns
    -------
    phot_table : `~astropy.table.Table`
        Astropy table with columns for flux, x/y coordinates of center,
        RA/dec coordinates of center, sky background per pixel,
        net flux, aperture and annulus radii used, and flux error.
    """

    # check that the outer radius is greater or equal the inner radius for annulus
    if inner_annulus >= outer_annulus:
        raise ValueError("outer_annulus must be greater than inner_annulus")

    # check that the annulus inner radius is greater or equal the aperture radius
    if aperture_radius >= inner_annulus:
        raise ValueError("inner_radius must be greater than aperture_radius")

    # Extract x,y coordinates from sources table, contstruct aperture and
    # annulus objects from coordinates, and peform aperture photometry
    coords = (sources['xcentroid'], sources['ycentroid'])
    apertures = CircularAperture(coords, aperture_radius)
    annulus = CircularAnnulus(coords, inner_annulus, outer_annulus)
    phot_table = aperture_photometry(ccd_image, apertures)
    phot_table_1 = aperture_photometry(ccd_image, annulus)

    # Obtain the local background/pixel and net flux between the aperture and annulus objects
    n_pix_ap = np.pi * (aperture_radius**2)
    n_pix_ann = np.pi * ((outer_annulus**2) - (inner_annulus**2))
    bkgd_pp = phot_table_1['aperture_sum'] / n_pix_ann
    net_flux = phot_table['aperture_sum'] -  (n_pix_ap * bkgd_pp)
    phot_table['background_per_pixel'] = bkgd_pp
    phot_table['net_flux'] = net_flux

    # Return a columns with the aperture radius and the inner/outer annulus radii
    phot_table['aperture_radius'] = np.ones(len(phot_table['aperture_sum'])) *  aperture_radius
    phot_table['inner_radius'] = np.ones(len(phot_table['aperture_sum'])) *  inner_annulus
    phot_table['outer_radius'] = np.ones(len(phot_table['aperture_sum'])) *  outer_annulus

    # Obtain RA/Dec coordinates and add them to table
    ra, dec = convert_pixel_wcs(ccd_image, coords[0], coords[1], 1)
    phot_table['RA_center'] = ra
    phot_table['Dec_center'] = dec

    # Obtain flux error and add column to return table
    noise = np.sqrt(gain*net_flux + n_pix_ap*(1 + (n_pix_ap/n_pix_ann))
			 *(gain*(bkgd_pp + N_dark_pp) + (N_R**2)))
    phot_table['aperture_sum_err'] = noise

    return phot_table



import inspect
import sys


def find_mod_objs(modname, onlylocals=False):
    """ Returns all the public attributes of a module referenced by name.

    .. note::
        The returned list *not* include subpackages or modules of
        `modname`,nor does it include private attributes (those that
        beginwith '_' or are not in `__all__`).

    Parameters
    ----------
    modname : str
        The name of the module to search.
    onlylocals : bool
        If True, only attributes that are either members of `modname` OR one of
        its modules or subpackages will be included.

    Returns
    -------
    localnames : list of str
        A list of the names of the attributes as they are named in the
        module `modname` .
    fqnames : list of str
        A list of the full qualified names of the attributes (e.g.,
        ``astropy.utils.misc.find_mod_objs``). For attributes that are
        simple variables, this is based on the local name, but for
        functions or classes it can be different if they are actually
        defined elsewhere and just referenced in `modname`.
    objs : list of objects
        A list of the actual attributes themselves (in the same order as
        the other arguments)

    """

    __import__(modname)
    mod = sys.modules[modname]

    if hasattr(mod, '__all__'):
        pkgitems = [(k, mod.__dict__[k]) for k in mod.__all__]
    else:
        pkgitems = [(k, mod.__dict__[k]) for k in dir(mod) if k[0] != '_']

    # filter out modules and pull the names and objs out
    ismodule = inspect.ismodule
    localnames = [k for k, v in pkgitems if not ismodule(v)]
    objs = [v for k, v in pkgitems if not ismodule(v)]

    # fully qualified names can be determined from the object's module
    fqnames = []
    for obj, lnm in zip(objs, localnames):
        if hasattr(obj, '__module__') and hasattr(obj, '__name__'):
            fqnames.append(obj.__module__ + '.' + obj.__name__)
        else:
            fqnames.append(modname + '.' + lnm)

    if onlylocals:
        valids = [fqn.startswith(modname) for fqn in fqnames]
        localnames = [e for i, e in enumerate(localnames) if valids[i]]
        fqnames = [e for i, e in enumerate(fqnames) if valids[i]]
        objs = [e for i, e in enumerate(objs) if valids[i]]

    return localnames, fqnames, objs
